#include "ehs/io/socket/Endpoint.h"
#include "ehs/io/socket/Comms.h"
#include "ehs/Encryption.h"
#include "ehs/system/CPU.h"

namespace ehs
{
	Endpoint::Endpoint()
		: owner(nullptr), disposition(EndDisp::UNKNOWN), status(Status::PENDING), arch(Architecture::UNKNOWN),
		hashId(0), nextSendId(0), nextRecvId(0), type(AddrType::IPV6), port(0), deltaDuration(0.0f),
		deltaRate(1.0f / 60.0f), timeout(0.0f), lastPing(0.0f), oldLatency(0.0f), latency(0.0f), queueSlot(0)
	{
		AddType("Endpoint");
	}

	Endpoint::Endpoint(Comms* owner, const EndDisp disposition, const Architecture arch, Str_8 id,
					   const AddrType type, Str_8 address, const UInt_16 port)
		: owner(owner), disposition(disposition), status(Status::ACTIVE), arch(arch), hashId(id.Hash_64()),
		id((Str_8&&)id), nextSendId(0), nextRecvId(0), type(type), address((Str_8&&)address), port(port),
		deltaDuration(0.0f), deltaRate(1.0f / 60.0f), timeout(0.0f), lastPing(0.0f), oldLatency(0.0f), latency(0.0f),
		queueSlot(0)
	{
		AddType("Endpoint");
	}

	Endpoint::Endpoint(Comms* owner, const AddrType type, Str_8 address, const UInt_16 port)
		: owner(owner), disposition(EndDisp::UNKNOWN), status(Status::PENDING), arch(Architecture::UNKNOWN), hashId(0),
		nextSendId(0), nextRecvId(0), type(type), address((Str_8&&)address), port(port), deltaDuration(0.0f),
		deltaRate(1.0f / 60.0f), timeout(0.0f), lastPing(0.0f), oldLatency(0.0f), latency(0.0f), queueSlot(0)
	{
		AddType("Endpoint");
	}

	Endpoint::Endpoint(const Endpoint& end)
		: BaseObj(end), owner(nullptr), disposition(end.disposition), status(Status::PENDING), arch(end.arch),
		hashId(end.hashId), id(end.id), nextSendId(0), nextRecvId(0), type(end.type), address(end.address),
		port(end.port), deltaDuration(0.0f), deltaRate(1.0f / 60.0f), timeout(0.0f), lastPing(0.0f),
		oldLatency(0.0f), latency(0.0f), queueSlot(0)
	{
	}

	Endpoint& Endpoint::operator=(const Endpoint& end)
	{
		if (this == &end)
			return *this;

		BaseObj::operator=(end);

		if (status == Status::PENDING)
		{
			disposition = end.disposition;
			status = end.status;
			arch = end.arch;
			hashId = end.hashId;
			id = end.id;
			nextSendId = end.nextSendId;
			sent = end.sent;
			nextRecvId = end.nextRecvId;
			received = end.received;
			type = end.type;
			address = end.address;
			port = end.port;
			deltaDuration = end.deltaDuration;
            deltaRate = end.deltaRate;
			timeout = end.timeout;
			lastPing = end.lastPing;
            oldLatency = end.oldLatency;
			latency = end.latency;
			queueSlot = end.queueSlot;
		}
		else
		{
			owner = nullptr;
			disposition = end.disposition;
			status = Status::PENDING;
			arch = end.arch;
			hashId = end.hashId;
			id = end.id;
			nextSendId = 0;
			sent = Vector<Insurance>();
			nextRecvId = 0;
			received = Vector<Fragments>();
			type = end.type;
			address = end.address;
			port = end.port;
			deltaDuration = 0.0f;
			deltaRate = 1.0f / 60.0f;
			timeout = 0.0f;
			lastPing = 0.0f;
            oldLatency = 0.0f;
			latency = 0.0f;
			queueSlot = 0;
		}

		return *this;
	}

	void Endpoint::Poll(const float delta)
	{
		SortReceived();

		if (deltaDuration >= deltaRate)
			deltaDuration = Math::Mod(deltaDuration, deltaRate);

		deltaDuration += delta;
		timeout += delta;
        latency += delta;

		if (sent.Size())
		{
            for (UInt_64 i = 0; i < sent.Size(); ++i)
            {
                sent[i].lastResend += delta;
                if (sent[i].lastResend >= owner->GetResendRate())
                {
                    Serializer result(Endianness::LE);
                    result.Write(sent[i].header);
                    result.WriteSer(sent[i].payload);

                    if (sent[i].header.encrypted)
                        Encryption::Encrypt_64(result.Size() - sizeof(bool), &result[sizeof(bool)]);

					owner->udp.Send(type, address, port, result, result.Size());

                    sent[i].lastResend = Math::Mod(sent[i].lastResend, owner->GetResendRate());
                }
            }
		}

		if (owner->GetDisposition() == EndDisp::SERVICE)
		{
			lastPing += delta;
			if (lastPing >= 1.0f)
				Ping(delta);
		}
	}

	EndDisp Endpoint::GetDisposition() const
	{
		return disposition;
	}

	void Endpoint::SetStatus(const Status newStatus)
	{
		status = newStatus;
	}

	Status Endpoint::GetStatus() const
	{
		return status;
	}

	Architecture Endpoint::GetArchitecture() const
	{
		return arch;
	}

	Str_8 Endpoint::GetId() const
	{
		return id;
	}

	UInt_64 Endpoint::GetHashId() const
	{
		return hashId;
	}

	UInt_64 Endpoint::GetNextSendId() const
	{
		return nextSendId;
	}

	void Endpoint::Send(const bool deltaLocked, const bool encrypted, const bool ensure, const UInt_64 sys,
                        const UInt_64 op, const Serializer<>& payload)
	{
		if (deltaLocked && deltaDuration < deltaRate)
			return;

		Header header = {
				encrypted,
				nextSendId++,
				1,
				0,
				ensure,
				owner->GetDisposition(),
				owner->GetHashId(),
				sys,
				op
		};

		if ((owner->GetLocalAddressType() == AddrType::IPV6 && payload.Size() > COMMS_IPV6_PAYLOAD) || (owner->GetLocalAddressType() == AddrType::IPV4 && payload.Size() > COMMS_IPV4_PAYLOAD))
		{
			Fragments frags = FragmentData(header, payload);
			for (UInt_64 i = 0; i < frags.Size(); ++i)
            {
                Header newHeader = frags.GetHeader();
                newHeader.fragment = i;

                Send(newHeader, frags[i]);
            }
		}
		else
		{
			Send(header, payload);
		}
	}

	void Endpoint::Send(const bool deltaLocked, const bool encrypted, const bool ensure, const Str_8& sys,
								const Str_8& op, const Serializer<>& payload)
	{
		Send(deltaLocked, encrypted, ensure, sys.Hash_64(), op.Hash_64(), payload);
	}

	void Endpoint::RemoveInsurance(const UInt_64 msgId, const UInt_64 fragment)
	{
		for (UInt_64 i = 0; i < sent.Size(); ++i)
		{
			if (sent[i].header.id == msgId && sent[i].header.fragment == fragment)
			{
				sent.Remove(i);
				break;
			}
		}

		timeout = 0.0f;
	}

	UInt_64 Endpoint::GetNextRecvId() const
	{
		return nextRecvId;
	}

	void Endpoint::AddReceived(const Header& header, const Serializer<>& payload)
	{
		Fragments* frags = nullptr;

		for (UInt_64 i = 0; i < received.Size(); ++i)
		{
			if (received[i].GetHeader().id == header.id)
			{
				if (received[i][header.fragment].Size())
					return;

				frags = &received[i];
                break;
			}
		}

		if (header.id > nextRecvId)
			nextRecvId = header.id + 1;

		if (frags)
			(*frags)[header.fragment] = payload;
		else
			received.Push({header, payload});

		timeout = 0.0f;
	}

	Vector<Fragments> Endpoint::GetReceived() const
	{
		return received;
	}

	Vector<Fragments>* Endpoint::GetReceived()
	{
		return &received;
	}

	Str_8 Endpoint::GetAddress() const
	{
		return address;
	}

	UInt_16 Endpoint::GetPort() const
	{
		return port;
	}

	void Endpoint::SetDeltaRate(const float newDeltaRate)
	{
		deltaRate = newDeltaRate;
	}

	float Endpoint::GetDeltaRate() const
	{
		return deltaRate;
	}

	float Endpoint::GetTimeout() const
	{
		return timeout;
	}

	float Endpoint::GetLastPing() const
	{
		return lastPing;
	}

	void Endpoint::Ping(const float delta)
	{
		Serializer payload(Endianness::LE);
		payload.Write(delta);

		Send(false, true, false, "Internal", "Ping", payload);
		lastPing = 0.0f;
		latency = 0.0f;
	}

	void Endpoint::Pong(const float delta)
	{
		Serializer payload(Endianness::LE);
		payload.Write(delta);

		Send(false, true, false, "Internal", "Pong", payload);

		timeout = 0.0f;
	}

	void Endpoint::SendLatency()
	{
        oldLatency = latency * 1000;

		Serializer sPayload(Endianness::LE);
		sPayload.Write(oldLatency);

		Send(false, true, false, "Internal", "Latency", sPayload);

        latency = 0.0f;
		timeout = 0.0f;
	}

	void Endpoint::SetLatency(const float newLatency)
	{
		oldLatency = newLatency;
	}

	float Endpoint::GetLatency() const
	{
		return oldLatency;
	}

	void Endpoint::SetQueueSlot(const UInt_64 slot)
	{
		queueSlot = slot;
	}

	UInt_64 Endpoint::GetQueueSlot() const
	{
		return queueSlot;
	}

	Fragments Endpoint::FragmentData(const Header& header, const Serializer<>& data)
	{
		Fragments result;

		if (owner->GetLocalAddressType() == AddrType::IPV6)
		{
            UInt_64 frags = data.Size() / COMMS_IPV6_PAYLOAD;
            if (data.Size() % COMMS_IPV6_PAYLOAD)
                ++frags;

			result = Fragments(header, frags);

			UInt_64 size = COMMS_IPV6_PAYLOAD;

			for (UInt_64 i = 0; i < result.Size(); ++i)
			{
				size = COMMS_IPV6_PAYLOAD;
				if (i == result.Size() - 1)
					size = data.Size() % COMMS_IPV6_PAYLOAD;

				result[i] = {data.GetEndianness(), &data[i * COMMS_IPV6_PAYLOAD], size};
			}
		}
		else if (owner->GetLocalAddressType() == AddrType::IPV4)
		{
            UInt_64 frags = data.Size() / COMMS_IPV4_PAYLOAD;
            if (data.Size() % COMMS_IPV4_PAYLOAD)
                ++frags;

			result = Fragments(header, frags);

			UInt_64 size = COMMS_IPV4_PAYLOAD;

			for (UInt_64 i = 0; i < result.Size(); ++i)
			{
				size = COMMS_IPV4_PAYLOAD;
				if (i == result.Size() - 1)
					size = data.Size() % COMMS_IPV4_PAYLOAD;

				result[i] = {data.GetEndianness(), &data[i * COMMS_IPV4_PAYLOAD], size};
			}
		}

		return result;
	}

	void Endpoint::Send(const Header& header, const Serializer<UInt_64>& payload)
	{
		Serializer result(Endianness::LE);
		result.Write(header);
		result.WriteSer(payload);

		if (header.encrypted)
			Encryption::Encrypt_64(result.Size() - sizeof(bool), &result[sizeof(bool)]);

		if (header.ensure)
			sent.Push({header, payload});

		owner->udp.Send(type, address, port, result, result.Size());
	}

	bool Endpoint::SortingNeeded() const
	{
		UInt_64 lastPacket = 0;

		for (UInt_64 i = 0; i < received.Size(); ++i)
		{
			if (received[i].GetHeader().id < lastPacket)
				return true;
			else
				lastPacket = received[i].GetHeader().id;
		}

		return false;
	}

	void Endpoint::SortReceived()
	{
		if (!SortingNeeded())
			return;

		Vector<Fragments> sorted(0, received.Stride());

		for (UInt_64 a = 0; a < received.Size(); ++a)
		{
			if (sorted.Size())
			{
				for (UInt_64 b = sorted.Size(); b; --b)
				{
					if (received[a].GetHeader().id > sorted[b - 1].GetHeader().id)
					{
						if (b == sorted.Size())
							sorted.Push(received[a]);
						else
							sorted.Insert(b, received[a]);

						break;
					}
					else
					{
						sorted.Insert(b - 1, received[a]);

						break;
					}
				}
			}
			else
			{
				sorted.Push(received[a]);
			}
		}

		received = sorted;
	}
}